/* 
 * Copyright (C) Shivaram Upadhyayula <shivaram.u@quadstor.com>
 */

#include "clstdefs.h"

atomic64_t ntransaction_id;

void
node_msg_free(struct core_node_msg *raw)
{
	__qs_free(raw, M_NODE_RMSG);
}

struct core_node_msg *
node_msg_alloc(int msg_len)
{
	struct core_node_msg *raw;

	raw = __qs_malloc(sizeof(*raw) + msg_len, M_NODE_RMSG, Q_WAITOK);
	return raw;
}

int
node_sock_write_page(sock_t *sock, pagestruct_t *page, int dxfer_len)
{
	int retval, offset = 0;

	while (1) {
		if (sock_state_error(sock))
			return -1;
		retval = sys_sock_write_page(sock, page, offset, dxfer_len);
		if (unlikely(retval < 0)) {
			debug_warn("write of %d bytes offset %d failed\n", dxfer_len, offset);
			node_sock_write_error(sock);
			return retval;
		}

		dxfer_len -= retval;
		if (!dxfer_len)
			return 0;
		offset += retval;
		node_sock_wait_for_write(sock);
		if (atomic_test_bit(SOCK_FLAGS_WRITE_WAIT, &sock->flags))
			return -1;
	}
	return 0;
}

int
node_sock_write_data(sock_t *sock, uint8_t *buffer, int dxfer_len)
{
	int retval;

	while (1) {
		if (sock_state_error(sock))
			return -1;

		retval = sys_sock_write(sock, buffer, dxfer_len);
		if (unlikely(retval < 0)) {
			debug_warn("write of %d bytes failed retval %d\n", dxfer_len, retval);
			node_sock_write_error(sock);
			return retval;
		}

		dxfer_len -= retval;
		if (!dxfer_len)
			return 0;
		buffer += retval;
		node_sock_wait_for_write(sock);
		if (atomic_test_bit(SOCK_FLAGS_WRITE_WAIT, &sock->flags))
			return -1;
	}
	return 0;

}

int 
node_sock_write(sock_t *sock, struct core_node_msg *raw)
{
	uint8_t *buffer = (uint8_t *)raw;
	int dxfer_len = (raw->dxfer_len + sizeof(*raw));

	return node_sock_write_data(sock, buffer, dxfer_len);
}

void
node_sock_read_error(sock_t *sock)
{
	atomic_set_bit(SOCK_FLAGS_READ_ERROR, &sock->flags);
	__chan_wakeup(&sock->sock_wait);
}

void
node_sock_write_error(sock_t *sock)
{
	atomic_set_bit(SOCK_FLAGS_READ_ERROR, &sock->flags);
	__chan_wakeup(&sock->sock_wait);
}

void
node_sock_finish(sock_t *sock)
{
	if (!sock_state_error(sock))
		put_comm_sock(sock);
}

void
node_msg_compute_csum(struct core_node_msg *raw)
{
	uint8_t *ptr = (uint8_t *)(raw);

//	raw->data_csum = csum32(raw->data, raw->dxfer_len);
	raw->csum = csum32((ptr + sizeof(uint32_t)), sizeof(*raw) - sizeof(uint32_t));
}

int
node_msg_csum_valid(struct core_node_msg *raw)
{
	uint8_t *ptr = (uint8_t *)(raw);
	uint32_t csum;

	csum = csum32((ptr + sizeof(uint32_t)), sizeof(*raw) - sizeof(uint32_t));
	return (csum == raw->csum);
}

int
node_send_msg(sock_t *sock, struct core_node_msg *raw)
{
	node_msg_compute_csum(raw);
	return node_sock_write(sock, raw);
}

int
node_sock_write_pgdata_list(sock_t *sock, struct pgdata *pglist, int pglist_cnt, uint32_t size)
{
	struct pgdata *pgdata;
	uint32_t min_len;
	int i, retval;

	for (i = 0; (i < pglist_cnt) && size; i++) {
		pgdata = &pglist[i];
		min_len = min_t(uint32_t, size, pgdata->pg_len);
		retval = node_sock_write_page(sock, pgdata->page, min_len);
		if (unlikely(retval != 0))
			return retval;
		size -= min_len;
	}
	return 0;
}

