/*
 * Copyright (C) Shivaram Upadhyayula <shivaram.u@quadstor.com>
 */
#include "clstdefs.h"
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include "missingdefs.h"
#include <commondefs.h>

#include "core_sock.c"

extern uma_t *node_sock_cache;

int
sock_state_error(sock_t *sock)
{
	return (atomic_test_bit(SOCK_FLAGS_READ_ERROR, &sock->flags) || atomic_test_bit(SOCK_FLAGS_IS_CLOSED, &sock->flags));
}

void
node_sock_start(sock_t *sock)
{
	sys_sock_nopush(sock, 1);
}

void
node_sock_end(sock_t *sock)
{
	sys_sock_nopush(sock, 0);
}

static void
node_sock_wait_for_read_timeout(sock_t *sock)
{
	wait_on_chan_timeout(sock->sock_wait, sys_sock_has_read_data(sock) || sock_state_error(sock), NODE_SOCK_TIMEOUT);
}

int
node_sock_read(sock_t *sock, void *buf, int dxfer_len)
{
	uint8_t *buffer = buf;
	int retval;

	while (1) {
		if (sock_state_error(sock))
			return -1;
		retval = sys_sock_read(sock, buffer, dxfer_len);
		if (unlikely(retval < 0)) {
			node_sock_read_error(sock);
			return retval;
		}

		dxfer_len -= retval;
		if (!dxfer_len)
			return 0;
		buffer += retval;
		node_sock_wait_for_read_timeout(sock);
	}
}

int
node_sock_read_nofail(sock_t *sock, void *buf, int dxfer_len)
{
	uint8_t *buffer = buf;
	int retval;

	while (1) {
		if (sock_state_error(sock))
			return -1;
		retval = sys_sock_read(sock, buffer, dxfer_len);
		if (unlikely(retval < 0)) {
			node_sock_read_error(sock);
			return retval;
		}

		dxfer_len -= retval;
		if (!dxfer_len)
			return 0;
		buffer += retval;
		node_sock_wait_for_read_timeout(sock);
		if (!sys_sock_has_read_data(sock)) {
			node_sock_read_error(sock);
			return -1;
		}
	}
}

void
node_sock_wait_for_write(sock_t *sock)
{
	unsigned long flags;

	chan_lock_intr(&sock->sock_wait, &flags);
	if (sys_sock_has_write_space(sock)) {
		chan_unlock_intr(&sock->sock_wait, &flags);
		return;
	}
	atomic_set_bit(SOCK_FLAGS_WRITE_WAIT, &sock->flags);
	chan_unlock_intr(&sock->sock_wait, &flags);

	wait_on_chan_timeout(sock->sock_wait, !atomic_test_bit(SOCK_FLAGS_WRITE_WAIT, &sock->flags) || sock_state_error(sock), NODE_SOCK_TIMEOUT);
}

static void
sys_sock_state_change(struct sock *sk)
{
	sock_t *sys_sock = sk->sk_user_data;
	unsigned long flags;

	if (!sys_sock)
		return;

	chan_lock_intr(&sys_sock->sock_wait, &flags);
	switch (sk->sk_state) {
	case TCP_ESTABLISHED:
		atomic_set_bit(SOCK_FLAGS_IS_CONNECTED, &sys_sock->flags);
		break;
	default:
		atomic_set_bit(SOCK_FLAGS_IS_CLOSED, &sys_sock->flags);
		break;
	}
	__chan_wakeup_unlocked(&sys_sock->sock_wait);
	chan_unlock_intr(&sys_sock->sock_wait, &flags);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,15,0))
static void sys_sock_data_ready(struct sock *sk)
#else
static void sys_sock_data_ready(struct sock *sk, int count)
#endif
{
	sock_t *sys_sock = sk->sk_user_data;

	if (!sys_sock)
		return;
	chan_wakeup(&sys_sock->sock_wait);
}

static void
sys_sock_write_space(struct sock *sk)
{
	sock_t *sys_sock = sk->sk_user_data;

	if (!sys_sock)
		return;

	if (sk_stream_wspace(sk) >= sk_stream_min_wspace(sk)) {
		chan_wakeup(&sys_sock->sock_wait);
	}
}


void
__node_sock_free(sock_t *sock)
{
	__uma_zfree(node_sock_cache, sock);
}

sock_t *
__node_sock_alloc(void *priv)
{
	sock_t *sock;

	sock = __uma_zalloc(node_sock_cache, M_WAITOK);
	sock->priv = priv;
	wait_chan_init(&sock->sock_wait, "node sock wait");
	return sock;
}

