#include "clstdefs.h"
#include <cluster_common.h>
#include <commondefs.h>

static int
ctio_read_pgdata_list(sock_t *sock, struct qsio_scsiio *ctio)
{
	struct pgdata *pglist, *pgdata;
	int i, offset = 0, min_len, retval;

	pglist = (struct pgdata *)(ctio->data_ptr);
	for (i = 0; i < ctio->pglist_cnt; i++) {
		pgdata = &pglist[i];
		min_len = ctio->dxfer_len - offset;
		if (min_len > pgdata->pg_len)
			min_len = pgdata->pg_len;
		retval = node_sock_read_nofail(sock, vm_pg_address(pgdata->page), min_len);
		if (unlikely(retval != 0))
			return -1;
		offset += min_len;
	}
	return 0;
}

int
core_msg_send_ctio_mmap(struct qsio_scsiio *ctio)
{
	struct tdisk *tdevice = ctio->ccb_h.tdevice;
	uint8_t *ptr = page_address(tdevice->ctio_page);
	struct core_node_msg *raw;
	struct mmap_entry *ret = NULL;
	struct ctio_req *req;
	struct ctio_rsp *rsp;

	raw = (struct core_node_msg *)(ptr);
	bzero(raw, sizeof(*req) + sizeof(*raw));
	raw->msg_cmd = NODE_MSG_CTIO_REQ;
	raw->dxfer_len = sizeof (*req);
	req = (struct ctio_req *)(raw->data);
	memcpy(req->cdb, ctio->cdb, 16);
	req->dxfer_len = ctio->dxfer_len;
	req->task_tag = ctio->task_tag;
	req->i_prt[0] = ctio->i_prt[0];
	req->i_prt[1] = ctio->i_prt[1];
	req->t_prt[0] = ctio->t_prt[0];
	req->t_prt[1] = ctio->t_prt[1];
	req->init_int = ctio->init_int;
	req->lun_id = tdevice->lun_id;
	req->r_prt = ctio->r_prt;
	req->tx_id = ctio->tx_id;
	if (ctio->init_int == TARGET_INT_ISCSI) {
		struct iscsi_priv *priv = &ctio->ccb_h.priv.ipriv;
		strcpy(req->init_name, priv->init_name);
	}

	raw->msg_id = node_transaction_id();
	node_msg_compute_csum(raw);

	tdevice_set_msg_pending(tdevice);
	ctio_free_data_nopage(ctio);

	if (raw->msg_status != NODE_STATUS_OK) {
		DEBUG_WARN_NEW("ctio %x msg status %x\n", ctio->cdb[0], raw->msg_status);
		memset(raw, 0, sizeof(*raw));
		return -1;
	}

	rsp = (struct ctio_rsp *)(raw->data);
	ctio->scsi_status = rsp->scsi_status;
	ctio->sense_len = rsp->sense_len;
	ctio->dxfer_len = rsp->ctio_dxfer_len;
	ptr += sizeof(*raw) + sizeof(*rsp);
	if (rsp->pglist_cnt && is_ctio_read(ctio->cdb[0])) {
		struct pgdata *pglist, *pgdata;
		struct pgdata_spec *spec;
		int spec_len, i;

		spec_len = (rsp->pglist_cnt * sizeof(*spec));
		if (rsp->dxfer_len != spec_len) {
			DEBUG_WARN_NEW("rsp dxfer len %d expected %d\n", rsp->dxfer_len, spec_len);
			return -1;
		}

		pglist = __qszalloc(sizeof(struct pgdata) * rsp->pglist_cnt, M_PGLIST, Q_WAITOK);
		spec = (struct pgdata_spec *)ptr;
		for (i = 0; i < rsp->pglist_cnt; i++, spec++) {
			pgdata = &pglist[i];
			pgdata->page = mmap_lookup_page((unsigned long)spec->addr, &ret);
			if (!pgdata->page) {
				DEBUG_WARN_NEW("Cannot lookup for addr %lu\n", spec->addr);
				__qs_free(pglist, M_PGLIST);
				return -1;
			}
			pgdata->pg_len = spec->pg_len;
			pgdata->pg_offset = spec->pg_offset;
		}
		ctio->pglist_cnt = rsp->pglist_cnt;
		ctio->data_ptr = (void *)pglist;
		ctio->tx_id = rsp->tx_id;
		ptr += spec_len;
	}
	else 
		DEBUG_BUG_ON(rsp->ctio_dxfer_len);
	if (rsp->sense_len) {
		int sense_offset = (ctio->init_int != TARGET_INT_ISCSI) ? 0 : 2;
		/* padding for iSCSI taken care by usr coredev */
		ctio->sense_data = zalloc(rsp->sense_len + sense_offset + 4, M_DEVBUF, M_WAITOK);
		memcpy(ctio->sense_data + sense_offset, ptr, ctio->sense_len);
		ctio->scsi_status = SCSI_STATUS_CHECK_COND;
		ctio->sense_len = rsp->sense_len;
	}
	memset(raw, 0, sizeof(*raw));
	return 0;
}

int
core_msg_send_ctio(struct qsio_scsiio *ctio)
{
	struct tdisk *tdevice = ctio->ccb_h.tdevice;
	sock_t *sock;
	struct ctio_req *req;
	struct mmap_entry *ret = NULL;
	struct ctio_rsp rsp;
	struct core_node_msg *raw, raw_rsp;
	struct pgdata *pglist;
	int error = -1, retval, pglist_cnt;

	sock = get_comm_sock();
	if (unlikely(!sock)) {
		debug_warn("cannot get a sock for cdb %x\n", ctio->cdb[0]);
		return -1;
	}

	raw = node_msg_alloc(sizeof(*req));
	bzero(raw, sizeof(*req) + sizeof(*raw));

	raw->msg_cmd = NODE_MSG_CTIO_REQ;
	raw->dxfer_len = sizeof (*req);
	req = (struct ctio_req *)(raw->data);
	memcpy(req->cdb, ctio->cdb, 16);
	req->dxfer_len = ctio->dxfer_len;
	req->task_tag = ctio->task_tag;
	req->i_prt[0] = ctio->i_prt[0];
	req->i_prt[1] = ctio->i_prt[1];
	req->t_prt[0] = ctio->t_prt[0];
	req->t_prt[1] = ctio->t_prt[1];
	req->init_int = ctio->init_int;
	req->lun_id = tdevice->lun_id;
	req->r_prt = ctio->r_prt;
	req->tx_id = ctio->tx_id;
	if (ctio->init_int == TARGET_INT_ISCSI) {
		struct iscsi_priv *priv = &ctio->ccb_h.priv.ipriv;
		strcpy(req->init_name, priv->init_name);
	}

	raw->msg_id = node_transaction_id();
	node_msg_compute_csum(raw);
	node_sock_start(sock);
	retval = node_sock_write(sock, raw);
	if (unlikely(retval != 0)) {
		DEBUG_WARN_NEW("Cannot write for cdb %x retval %d\n", ctio->cdb[0], retval);
		node_sock_end(sock);
		goto out;
	}
	retval = 0;
	if (is_ctio_write(ctio->cdb[0]))
		retval = 0;
	else if (ctio->pglist_cnt)
		retval = node_sock_write_pgdata_list(sock, (struct pgdata *)ctio->data_ptr, ctio->pglist_cnt, ctio->dxfer_len);
	else if (ctio->dxfer_len) {
		retval = node_sock_write_data(sock, ctio->data_ptr, ctio->dxfer_len);
		if (retval != 0)
			DEBUG_INFO_NEW("failed for cdb %x dxfer len %d\n", ctio->cdb[0], ctio->dxfer_len);
	}
	node_sock_end(sock);
	if (retval != 0)
		goto out;

	if (is_ctio_write(ctio->cdb[0]))
		ctio_free_data_nopage(ctio);
	else
		ctio_free_data(ctio);
	retval = node_sock_read(sock, &raw_rsp, sizeof(raw_rsp));
	if (unlikely(retval != 0))
		goto out;

	if (raw_rsp.msg_status != NODE_STATUS_OK)
		goto out;

	if (raw_rsp.dxfer_len < sizeof(rsp))
		goto out;

	retval = node_sock_read_nofail(sock, &rsp, sizeof(rsp));
	if (unlikely(retval != 0))
		goto out;

	ctio->scsi_status = rsp.scsi_status;
	ctio->sense_len = rsp.sense_len;
	if (rsp.pglist_cnt && is_ctio_read(ctio->cdb[0])) {
		struct pgdata *pglist, *pgdata;
		struct pgdata_spec *spec;
		uint8_t *spec_ptr = NULL;
		int spec_len, i;

		spec_len = (rsp.pglist_cnt * sizeof(*spec));
		if (rsp.dxfer_len != spec_len) {
			DEBUG_WARN_NEW("rsp dxfer len %d expected %d\n", raw_rsp.dxfer_len, spec_len);
			goto out;
		}
		spec_ptr = __qs_malloc(spec_len, M_MMAP, Q_WAITOK);

		retval = node_sock_read(sock, spec_ptr, spec_len);
		if (unlikely(retval != 0)) {
			free(spec_ptr, M_MMAP);
			goto out;
		}

		pglist = __qszalloc(sizeof(struct pgdata) * rsp.pglist_cnt, M_PGLIST, Q_WAITOK);
		spec = (struct pgdata_spec *)spec_ptr;
		for (i = 0; i < rsp.pglist_cnt; i++, spec++) {
			pgdata = &pglist[i];
			pgdata->page = mmap_lookup_page((unsigned long)spec->addr, &ret);
			if (!pgdata->page) {
				DEBUG_WARN_NEW("Cannot lookup for addr %lu\n", spec->addr);
				free(spec_ptr, M_MMAP);
				__qs_free(pglist, M_PGLIST);
				goto out;
			}
			pgdata->pg_len = spec->pg_len;
			pgdata->pg_offset = spec->pg_offset;
		}
		free(spec_ptr, M_MMAP);
		ctio->dxfer_len = rsp.dxfer_len;
		ctio->dxfer_len = rsp.ctio_dxfer_len;
		ctio->pglist_cnt = rsp.pglist_cnt;
		ctio->data_ptr = (void *)pglist;
		ctio->tx_id = rsp.tx_id;
	}
	else if (rsp.dxfer_len) {
		pglist_cnt = 0;
		if (rsp.dxfer_len < 65536)
			ctio_allocate_buffer(ctio, rsp.dxfer_len, 0); 
		if (!ctio->data_ptr) {
			ctio->dxfer_len = 0;
			pglist = pgdata_allocate(rsp.dxfer_len, 1, &pglist_cnt, 0, 1); 
			if (!pglist) {
				ctio->dxfer_len = 0;
				ctio->pglist_cnt = 0;
				goto out;
			}
			ctio->data_ptr = (void *)pglist;
		}
		ctio->pglist_cnt = pglist_cnt;
		ctio->dxfer_len = rsp.dxfer_len;

		if (ctio->pglist_cnt) {
			retval = ctio_read_pgdata_list(sock, ctio);
			if (unlikely(retval != 0)) {
				ctio_free_data(ctio);
				goto out;
			}
		}
		else if (ctio->dxfer_len) {
			retval = node_sock_read_nofail(sock, ctio->data_ptr, ctio->dxfer_len);
			if (unlikely(retval != 0)) {
				ctio_free_data(ctio);
				goto out;
			}
		}
	}

	if (rsp.sense_len) {
		int sense_offset = (ctio->init_int != TARGET_INT_ISCSI) ? 0 : 2;
		/* padding for iSCSI taken care by usr coredev */
		ctio->sense_data = zalloc(rsp.sense_len + sense_offset + 4, M_DEVBUF, M_WAITOK);
		retval = node_sock_read_nofail(sock, ctio->sense_data+sense_offset, ctio->sense_len);
		if (unlikely(retval != 0)) {
			ctio_free_data(ctio);
			ctio_free_sense(ctio);
			goto out;
		}
		ctio->scsi_status = SCSI_STATUS_CHECK_COND;
		ctio->sense_len = rsp.sense_len;
	}
	error = 0;
out:
	node_sock_finish(sock);
	node_msg_free(raw);
	return error;
}
