/* Socket related stuff */

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,15,0))
static void sys_sock_data_ready(struct sock *sk);
#else
static void sys_sock_data_ready(struct sock *sk, int count);
#endif
static void sys_sock_write_space(struct sock *sk);
static void sys_sock_state_change(struct sock *sk);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
int
sys_sock_has_read_data(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;
	int retval;

	if (sys_sock->listen)
		return inet_csk_listen_poll(sock->sk);
	if (sys_sock->usock)
		retval = unix_inq_len(sock->sk);
	else
		retval = tcp_peek_len(sock);
	return retval;
}
#else
int
sys_sock_has_read_data(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;
	mm_segment_t oldfs;
	int avail = 0;
	int retval;

	if (sys_sock->listen)
		return inet_csk_listen_poll(sock->sk);
	oldfs = get_fs();
	set_fs(KERNEL_DS);
	retval = sock->ops->ioctl(sock, SIOCINQ, (unsigned long)&avail);
	set_fs(oldfs);
	return (retval >= 0) ? avail : 0;
}
#endif

int
sys_sock_has_write_space(sock_t *sys_sock)
{
	return (sk_stream_wspace(sys_sock->sock->sk) >= sk_stream_min_wspace(sys_sock->sock->sk));
}

int
sys_sock_is_connected(sock_t *sys_sock)
{
	return (sys_sock->sock->sk->sk_state == TCP_ESTABLISHED);
}

int
sys_sock_read(sock_t *sys_sock, void *buf, int len)
{
	struct msghdr msg;
	struct kvec iov;
	int retval;

	iov.iov_base = buf;
	iov.iov_len = len;

	memset(&msg, 0, sizeof(msg));
	msg.msg_flags = MSG_DONTWAIT | MSG_NOSIGNAL;
	retval = kernel_recvmsg(sys_sock->sock, &msg, &iov, 1, len, MSG_DONTWAIT | MSG_NOSIGNAL);
	if (retval > 0)
		return retval;
	else if (retval == -EAGAIN || retval == -EINTR)
		return 0;
	else
		return retval;
}

void
sys_sock_free(sock_t *sys_sock)
{
	sock_release(sys_sock->sock);
	__node_sock_free(sys_sock);
}

static void
sock_deactivate(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;

	write_lock_bh(&sock->sk->sk_callback_lock);

	/* Restore callbacks */
	if (sys_sock->state_change) {
		sock->sk->sk_state_change = sys_sock->state_change;
		sys_sock->state_change = NULL;
	}

	if (sys_sock->data_ready) {
		sock->sk->sk_data_ready = sys_sock->data_ready;
		sys_sock->data_ready = NULL;
	}

	if (sys_sock->write_space) {
		sock->sk->sk_write_space = sys_sock->write_space;
		sys_sock->write_space = NULL;
	}

	sock->sk->sk_user_data = NULL;
	write_unlock_bh(&sock->sk->sk_callback_lock);
}

int
sys_sock_shutdown(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;
	return kernel_sock_shutdown(sock, SHUT_RDWR);
}

void
sys_sock_close(sock_t *sys_sock, int linger)
{
	struct socket *sock = sys_sock->sock;

	if (!linger)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
		sock_no_linger(sock->sk);
#else
		kernel_setsockopt(sock, SOL_SOCKET, SO_LINGER, (void *)&linger, sizeof(linger));
#endif

	kernel_sock_shutdown(sock, SHUT_RDWR);
	sock_deactivate(sys_sock);
}

int
sys_sock_write_page(sock_t *sys_sock, pagestruct_t *page, int offset, int len)
{
	int retval;

	retval =  qs_sock_write_page(sys_sock->sock, page, offset, len, 0);
	if (retval > 0)
		return retval;
	else if (retval == -EAGAIN || retval == -EINTR)
		return 0;
	else
		return retval;
}

int
sys_sock_write(sock_t *sys_sock, void *buf, int len)
{
	int retval;

	retval = qs_sock_write(sys_sock->sock, buf, len, MSG_DONTWAIT | MSG_NOSIGNAL);
	if (retval > 0)
		return retval;
	else if (retval == -EAGAIN || retval == -EINTR)
		return 0;
	else
		return retval;
}

static sock_t *
sys_sock_create_lite(void *priv)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = __node_sock_alloc(priv);

	retval = sock_create_lite(AF_INET, SOCK_STREAM, IPPROTO_TCP, &sys_sock->sock);
	if (unlikely(retval < 0)) {
		__node_sock_free(sys_sock);
		return NULL;
	}
	return sys_sock;
}

sock_t *
__sys_sock_create(void *priv, int domain)
{
	sock_t *sys_sock;
	int retval;

	sys_sock = __node_sock_alloc(priv);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,2,0))
	retval = sock_create_kern(&init_net, domain, SOCK_STREAM, domain == AF_LOCAL ? 0: IPPROTO_TCP, &sys_sock->sock);
#else
	retval = sock_create_kern(domain, SOCK_STREAM, domain == AF_LOCAL ? 0: IPPROTO_TCP, &sys_sock->sock);
#endif
	if (unlikely(retval < 0)) {
		DEBUG_WARN_NEW("Cannot create sock retval %d\n", retval);
		__node_sock_free(sys_sock);
		return NULL;
	}
	sys_sock->priv = priv;
	return sys_sock;
}

sock_t *
sys_sock_create(void *priv)
{
	return __sys_sock_create(priv, AF_INET);
}

sock_t *
sys_sock_create_local(void *priv)
{
	return __sys_sock_create(priv, AF_LOCAL);
}

void
sock_activate(sock_t *sys_sock)
{
	struct socket *sock = sys_sock->sock;

	write_lock_bh(&sock->sk->sk_callback_lock);
	sock->sk->sk_user_data = sys_sock;
	sock->sk->sk_allocation = GFP_NOFS;

	/* Save callbacks */
	sys_sock->state_change = sock->sk->sk_state_change;
	sys_sock->data_ready = sock->sk->sk_data_ready;
	sys_sock->write_space = sock->sk->sk_write_space;

	/* Set new callbacks */
	sock->sk->sk_state_change = sys_sock_state_change;
	sock->sk->sk_write_space = sys_sock_write_space;
	sock->sk->sk_data_ready = sys_sock_data_ready;
	write_unlock_bh(&sock->sk->sk_callback_lock);
}

void
sys_sock_nopush(sock_t *sys_sock, int set)
{
	if (sys_sock->usock)
		return;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
	tcp_sock_set_cork(sys_sock->sock->sk, set);
#else
	kernel_setsockopt(sys_sock->sock, SOL_TCP, TCP_CORK, (void *)&set, sizeof(set));
#endif
}

sock_t *
sys_sock_accept(sock_t *sys_sock, void *priv, int *error, uint32_t *ipaddr)
{
	sock_t *new_syssock;
	struct socket *sock = sys_sock->sock;
	struct socket *newsock;
	struct sockaddr_in *sinaddr;
	int retval;

	new_syssock = sys_sock_create_lite(priv);
	new_syssock->usock = sys_sock->usock;
	newsock = new_syssock->sock;
	newsock->type = sock->type;
	newsock->ops = sock->ops;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,12,0))
	retval = sock->ops->accept(sock, newsock, O_NONBLOCK, false);
#else
	retval = sock->ops->accept(sock, newsock, O_NONBLOCK);
#endif
	if (retval < 0) {
		*error = retval;
		sock_release(newsock);
		__node_sock_free(new_syssock);
		return NULL;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4,18,0))
	retval = newsock->ops->getname(newsock, (struct sockaddr *)&new_syssock->saddr, 2);
#else
	retval = newsock->ops->getname(newsock, (struct sockaddr *)&new_syssock->saddr, &new_syssock->saddr_len, 2);
#endif
	if (retval < 0) {
		*error = retval;
		sock_release(newsock);
		__node_sock_free(new_syssock);
		return NULL;
        }

	sock_activate(new_syssock);
	if (newsock->sk->sk_state == TCP_ESTABLISHED) {
		sys_sock_state_change(newsock->sk);
	}

	if (!new_syssock->usock) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
		tcp_sock_set_nodelay(newsock->sk);
#else
		int disabled = 1;
		kernel_setsockopt(newsock, SOL_TCP, TCP_NODELAY, (void *)&disabled, sizeof(disabled));
#endif
	}
	sinaddr = (struct sockaddr_in *)(&new_syssock->saddr);
	*ipaddr = sinaddr->sin_addr.s_addr;
	*error = 0;
	return new_syssock;
}

int
sys_sock_bind(sock_t *sys_sock, uint32_t addr, uint16_t port)
{
	struct socket *sock = sys_sock->sock;
	struct sockaddr_in saddr_in;
	int retval;

	memset(&saddr_in, 0, sizeof(saddr_in));
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = htons(port);
	saddr_in.sin_addr.s_addr = addr; 

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
	sock_set_reuseaddr(sock->sk);
#else
	{
		int reuse = 1;
		kernel_setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&reuse, sizeof(reuse));
	}
#endif
	retval = sock->ops->bind(sock, (struct sockaddr *)&saddr_in, sizeof(saddr_in));
	if (retval < 0)
		return retval;

	retval = sock->ops->listen(sock, 1024);
	if (retval < 0)
		return retval;

	sys_sock->listen = 1;
	sock_activate(sys_sock);
	return 0;
}

int
sys_sock_connect_local(sock_t *sys_sock, char *path)
{
	struct socket *sock = sys_sock->sock;
	struct sockaddr_un un_addr;
	size_t size;
	int retval;

	memset(&un_addr, 0, sizeof(un_addr));
	un_addr.sun_family = AF_LOCAL;
	strcpy((char *)un_addr.sun_path, path);
	sys_sock->usock = 1;

	sock_activate(sys_sock);
	size = (offsetof (struct sockaddr_un, sun_path) + strlen (un_addr.sun_path));
	retval = sock->ops->connect(sock, (struct sockaddr *)&un_addr, size, O_NONBLOCK);
	if (retval != 0 && retval != -EINPROGRESS)
		return retval;
	return 0;
}

int
sys_sock_connect(sock_t *sys_sock, uint32_t addr, uint32_t local_addr, uint16_t port)
{
	struct socket *sock = sys_sock->sock;
	struct sockaddr_in saddr_in;
	int retval;

	if (!local_addr || (local_addr == addr))
		goto skip_bind;

	memset(&saddr_in, 0, sizeof(saddr_in));
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_addr.s_addr = local_addr; 

	retval = sock->ops->bind(sock, (struct sockaddr *)&saddr_in, sizeof(saddr_in));
	if (retval < 0)
		return retval;

skip_bind:
	memset(&saddr_in, 0, sizeof(saddr_in));
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = htons(port);
	saddr_in.sin_addr.s_addr = addr; 
	sock_activate(sys_sock);
	retval = sock->ops->connect(sock, (struct sockaddr *)&saddr_in, sizeof(saddr_in), O_NONBLOCK);
	if (retval != 0 && retval != -EINPROGRESS)
		return retval;

	if (!sys_sock->usock) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0))
		tcp_sock_set_nodelay(sock->sk);
#else
		int disabled = 1;
		kernel_setsockopt(sock, SOL_TCP, TCP_NODELAY, (void *)&disabled, sizeof(disabled));
#endif
	}
	return 0;
}

static atomic_t vmap_count;

void *
vm_pg_map(pagestruct_t **pp, int pg_count)
{
	void *addr;
	int tries = 0;

	atomic_inc(&vmap_count);
	if (atomic_read(&vmap_count) > 2048) {
		atomic_set(&vmap_count, 0);
		vm_unmap_aliases(); /* Unmap aliases periodically */
	}

retry:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,8,0)) ||  (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 5)
	addr = vm_map_ram(pp, pg_count, -1);
#else
	addr = vm_map_ram(pp, pg_count, -1, PAGE_KERNEL);
#endif
	if (addr || ++tries > 1)
		return addr;
	vm_unmap_aliases();
	goto retry;
}

void
vm_pg_unmap(void *maddr, int pg_count)
{
	vm_unmap_ram(maddr, pg_count);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,11,0))
int
validate_block_device(const char *devpath,  iodev_t *b_dev)
{
	dev_t lookup;
	int error;

	error = lookup_bdev(devpath, &lookup);
	if (error) {
		DEBUG_WARN_NEW("Cannot lookup %s\n", devpath);
		return -1;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6,10,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR == 9 && RHEL_MINOR >= 5)
	if (file_bdev(b_dev)->bd_dev == lookup)
#else
	if (b_dev->bd_dev == lookup)
#endif
		return 1;
	else {
		DEBUG_INFO_NEW("devpath %s changed\n", devpath);
		return 0;
	}
}
#else
int
validate_block_device(const char *devpath,  iodev_t *b_dev)
{
	iodev_t *lookup;

	lookup = lookup_bdev(devpath);
	if (unlikely(IS_ERR(b_dev))) {
		DEBUG_WARN_NEW("Cannot lookup %s\n", devpath);
		return -1;
	}
	bdput(lookup);
	DEBUG_INFO_NEW("devpath %s lookup %p b_dev %p\n", devpath, lookup, b_dev);
	if (lookup == b_dev)
		return 1;
	else {
		DEBUG_INFO_NEW("devpath %s changed to lookup %p from b_dev %p\n", devpath, lookup, b_dev);
		return 0;
	}
}
#endif
