/* 
 * Copyright (C) Shivaram Upadhyayula <shivaram.u@quadstor.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301, USA.
 */

#include <linuxdefs.h>
#include <exportdefs.h>
#include <missingdefs.h>
#include "qla_def.h"
#include "qla_target.h"
#include "fcq.h"
#include "qla_sc.h"

static int
cmd_r_prt(struct fcbridge *fcbridge)
{
	return (FC_RPORT_START + fcbridge->id);
}

static inline struct fc_cmd *
get_next_cmd(struct fcq *fcq)
{
	struct fc_cmd *cmd;
	unsigned long flags = 0;

	spin_lock_irqsave(&fcq->fcq_lock, flags);
	cmd = TAILQ_FIRST(&fcq->pending_queue);
	if (cmd)
		TAILQ_REMOVE(&fcq->pending_queue, cmd, q_list); 
	spin_unlock_irqrestore(&fcq->fcq_lock, flags);
	return cmd;
}

extern struct qs_interface_cbs icbs;
extern sx_t itf_lock;
extern atomic_t alloced_cmds;

static inline void 
process_ctio(struct fcq *fcq, struct fc_cmd *fc_cmd)
{
	struct qla_tgt_cmd *cmd = fc_cmd->tgt_cmd;
	struct qla_tgt *tgt;
	struct lport *lport;
	struct fcbridge *fcbridge;
	struct qsio_scsiio *ctio;
	struct qpriv *priv;
	uint32_t target_lun;

	tgt = cmd->sess->tgt;
	lport = tgt->vha->vha_tgt.target_lport_ptr;
	fcbridge = lport->fcbridge;
	if (fc_cmd->need_free) {
		struct fc_cmd *fc_cmd = (struct fc_cmd *)cmd->se_cmd.priv;
		struct qsio_scsiio *ctio = (struct qsio_scsiio *)fc_cmd->ccb;

		if (ctio->ccb_h.flags & QSIO_DIR_OUT && !ctio->local_pool && !(ctio->ccb_h.flags & QSIO_DONE_SEND_STATUS)) {
			struct ccb_list ctio_list;
	
			STAILQ_INIT(&ctio_list);
			(*icbs.device_remove_ctio)(ctio, &ctio_list);
			(*icbs.device_queue_ctio_list)(&ctio_list);
		}
		ctio->end_timestamp = get_jiffies_64();
		__ctio_free_all(ctio, ctio->local_pool);
		qla_sc_session_remove_cmd(fc_cmd);
		cmd_sg_free(cmd);
		qlt_free_cmd(cmd);
		return;
	}

	if (fc_cmd->ccb)
	{
		struct fc_cmd *fc_cmd = (struct fc_cmd *)cmd->se_cmd.priv;
		struct qsio_scsiio *ctio = (struct qsio_scsiio *)fc_cmd->ccb;
		struct tdevice *device = ctio->ccb_h.tdevice;

		if (cmd->aborted || !cmd->write_data_transferred) {
			qla_handle_incomp_ctio(ctio);
			return;
		}

		if (device) {
			if (ctio->istate)
				(*icbs.device_queue_ctio_direct) ((struct qsio_hdr *)ctio);
			else
				fcbridge_insert_cmd(ctio);
			return;
		}
		/* ctio completed (data recevied) */
		fcbridge_proc_cmd(fcbridge, ctio);
		return;
	}

	target_lun = cmd->unpacked_lun;
	sx_xlock(&itf_lock);
	if (atomic_read(&icbs.itf_enabled) && target_lun)
	{
		atomic_inc(&alloced_cmds);
		ctio = (*icbs.ctio_new)(Q_WAITOK);
	}
	else
	{
		ctio = __local_ctio_new(M_WAITOK);
		ctio->local_pool = 1;
	}
	sx_xunlock(&itf_lock);

	ctio->start_timestamp = cmd->jiffies_at_alloc;
	ctio->i_prt[0] = wwn_to_u64(cmd->sess->port_name);
	ctio->t_prt[0] = wwn_to_u64(tgt->vha->port_name);
	ctio->t_prt[1] = wwn_to_u64(tgt->vha->node_name);
	ctio->r_prt = cmd_r_prt(fcbridge);
	ctio->init_int = TARGET_INT_FC;
	memcpy(ctio->cdb, fc_cmd->cdb, 16);
	ctio->ccb_h.flags = QSIO_DIR_OUT;
	ctio->ccb_h.flags |= QSIO_TYPE_CTIO;
	ctio->ccb_h.queue_fn = qla_end_ccb;
	ctio->ccb_h.target_lun = target_lun;
	ctio->task_attr = fc_cmd->task_attr;
	ctio->task_tag = cmd->se_cmd.tag;
	fc_cmd->ccb = ctio;

	priv = &ctio->ccb_h.priv.qpriv;
	priv->qcmd = fc_cmd;
	priv->fcbridge = fcbridge;
	priv->fcq = fcq;
	fcbridge_proc_cmd(fcbridge, ctio);
}

static void
process_inot(struct fcq *fcq, struct fc_cmd *fc_cmd)
{
	struct qla_tgt_mgmt_cmd *cmd = fc_cmd->tgt_cmd;
	struct qla_tgt *tgt = cmd->sess->tgt;
	struct lport *lport = tgt->vha->vha_tgt.target_lport_ptr;
	struct fcbridge *fcbridge = lport->fcbridge;
	struct qsio_immed_notify *notify;
	struct qpriv *priv;

	notify = kzalloc(sizeof(struct qsio_immed_notify), GFP_KERNEL|__GFP_NOFAIL);

	notify->i_prt[0] = wwn_to_u64(cmd->sess->port_name);
	notify->t_prt[0] = wwn_to_u64(tgt->vha->port_name);
	notify->r_prt = cmd_r_prt(fcbridge);
	notify->init_int = TARGET_INT_FC;
	notify->fn = fc_cmd->notify_fn;
	notify->ccb_h.target_lun = cmd->unpacked_lun;
	notify->ccb_h.flags = QSIO_TYPE_NOTIFY;
	notify->ccb_h.queue_fn = qla_end_ccb;
	notify->task_tag = cmd->se_cmd.tag;

	priv = &notify->ccb_h.priv.qpriv;
	priv->qcmd = fc_cmd;
	priv->fcbridge = fcbridge;
	fcbridge_task_mgmt(fcbridge, notify);
}

static inline void 
process_cmd(struct fcq *fcq, struct fc_cmd *cmd_h)
{
	if (cmd_h->type == QLA_HDR_TYPE_CTIO)
		process_ctio(fcq, cmd_h);
	else if (cmd_h->type == QLA_HDR_TYPE_NOTIFY)
		process_inot(fcq, cmd_h);
	else {
		DEBUG_BUG_ON(1);
		DEBUG_WARN_NEW("Invalid cmd type %x\n", cmd_h->type);
	}
}

/* process_queue returns only after draining the queue */
static inline void
process_queue(struct fcq *fcq)
{
	struct fc_cmd *cmd;

	while ((cmd = get_next_cmd(fcq)) != NULL)
	{
		/* process the commands.  */
		process_cmd(fcq, cmd);
	}
}

extern mempool_t *q2t_pool;

static int
fcq_thread(void *data)
{
	struct fcq *fcq = data;

	set_user_nice(current, -20);

	__set_current_state(TASK_RUNNING);

	for (;;)
	{
		wait_event_interruptible(fcq->fcq_wait, !TAILQ_EMPTY(&fcq->pending_queue) || kthread_should_stop());

		if (unlikely(kthread_should_stop()))
		{
			break;
		}

		process_queue(fcq);
	}
	return 0;
}


void
fcq_exit(struct fcq *fcq)
{
	int err;

	err = kthread_stop(fcq->task);

	if (err)
	{
		DEBUG_WARN_NEW("Shutting down fcq thread failed\n");
		return;
	}	

	kfree(fcq);
}

struct fcq *
fcq_init(char *namefmt, int id)
{
	struct fcq *fcq;
	struct task_struct *task;

	fcq = kzalloc(sizeof(struct fcq), GFP_KERNEL);
	if (unlikely(!fcq))
		return NULL;

	init_waitqueue_head(&fcq->fcq_wait);
	TAILQ_INIT(&fcq->pending_queue);
	spin_lock_init(&fcq->fcq_lock);
	task = kthread_run(fcq_thread, fcq, namefmt, id);

	if (IS_ERR(task)) {
		kfree(fcq);
		return NULL;
	}
	fcq->task = task;
	return fcq;
}
