/* 
 * Copyright (C) Shivaram Upadhyayula <shivaram.u@quadstor.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * Version 2 as published by the Free Software Foundation
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, 
 * Boston, MA  02110-1301, USA.
 */

#include <linuxdefs.h>
#include <exportdefs.h>
#include <missingdefs.h>
#include <commondefs.h>
#include <ioctldefs.h>
#include "qla_def.h"
#include "qla_target.h"
#include "qla_sc.h"
#include "fcq.h"

static struct kmem_cache *fc_sess_cache;
static struct kmem_cache *fc_cmd_cache;
static mempool_t *fc_cmd_pool;
static struct kmem_cache *tgt_cmd_cache;
static struct kmem_cache *lport_cache;
static struct kmem_cache *se_session_cache;
static struct qla_tgt_func_tmpl qla_sc_template;

static struct fc_sess *
fc_sess_new(struct lport *lport)
{
	struct fc_sess *fc_sess;

	fc_sess = kmem_cache_zalloc(fc_sess_cache, GFP_ATOMIC);
	if (!fc_sess)
		return NULL;

	spin_lock_init(&fc_sess->sess_lock);
	TAILQ_INIT(&fc_sess->cmd_list);
	return fc_sess;
}

static void
fc_sess_remove(struct lport *lport, struct fc_sess *fc_sess)
{
	unsigned long flags;

	spin_lock_irqsave(&lport->lport_lock, flags);
	if (fc_sess->active) {
		DEBUG_WARN_NEW("lport %d fc sess still active loop_id %d key %x\n", lport->id, fc_sess->loop_id, fc_sess->key);
		lport->loop_sess_list[fc_sess->loop_id] = NULL;
		TAILQ_REMOVE(&lport->active_sess_list, fc_sess, a_list);
	}
	TAILQ_REMOVE(&lport->sess_list, fc_sess, s_list);
	spin_unlock_irqrestore(&lport->lport_lock, flags);

	DEBUG_BUG_ON(!TAILQ_EMPTY(&fc_sess->cmd_list));
	kmem_cache_free(fc_sess_cache, fc_sess);
}

static void
fc_sess_add_cmd(struct fc_sess *fc_sess, struct fc_cmd *fc_cmd)
{
	unsigned long flags;

	spin_lock_irqsave(&fc_sess->sess_lock, flags);
	TAILQ_INSERT_TAIL(&fc_sess->cmd_list, fc_cmd, f_list);
	atomic_inc(&fc_sess->pending_cmds);
	spin_unlock_irqrestore(&fc_sess->sess_lock, flags);
}

static void
fc_sess_remove_cmd(struct fc_sess *fc_sess, struct fc_cmd *fc_cmd)
{
	unsigned long flags;

	spin_lock_irqsave(&fc_sess->sess_lock, flags);
	TAILQ_REMOVE(&fc_sess->cmd_list, fc_cmd, f_list);
	atomic_dec(&fc_sess->pending_cmds);
	spin_unlock_irqrestore(&fc_sess->sess_lock, flags);
}

struct lport_list lport_list = TAILQ_HEAD_INITIALIZER(lport_list);
DEFINE_SPINLOCK(lport_list_lock);

void
ctio_sglist_map(struct qsio_scsiio *ctio, struct qla_tgt_cmd *cmd)
{
	struct scatterlist *sglist;
	struct pgdata *pglist;
	int i;
	int dxfer_len = ctio->dxfer_len;

	DEBUG_BUG_ON(cmd->sg);
	while ((sglist = kmalloc(ctio->pglist_cnt * sizeof(*sglist), GFP_KERNEL)) == NULL)
		msleep(10);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
	sg_init_table(sglist, ctio->pglist_cnt);
#endif

	pglist = (struct pgdata *)ctio->data_ptr;
	for (i = 0; i < ctio->pglist_cnt; i++)
	{
		struct pgdata *pgtmp = &pglist[i];
		struct scatterlist *sgtmp = &sglist[i];
		int min_len;

		min_len = min_t(int, dxfer_len, pgtmp->pg_len); 
		dxfer_len -= min_len;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
		sg_set_page(sgtmp, pgtmp->page, min_len, pgtmp->pg_offset);
#else
		sgtmp->page = pgtmp->page;
		sgtmp->offset = pgtmp->pg_offset;
		sgtmp->length = min_len;
#endif
	}

	cmd->sg = cmd->se_cmd.t_data_sg = sglist;
	cmd->sg_cnt = cmd->se_cmd.t_data_nents = ctio->pglist_cnt;
}

static void
ctio_buffer_map(struct qsio_scsiio *ctio, struct qla_tgt_cmd *cmd)
{
	struct scatterlist *sgtmp;

	sgtmp = kmalloc(sizeof(struct scatterlist), GFP_KERNEL|__GFP_NOFAIL);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
	sg_init_table(sgtmp, 1);
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
	sg_set_page(sgtmp, virt_to_page(ctio->data_ptr), ctio->dxfer_len, page_offset(ctio->data_ptr));
#else
	sgtmp->page = virt_to_page(ctio->data_ptr);
	sgtmp->offset = page_offset(ctio->data_ptr);
	sgtmp->length = ctio->dxfer_len;
#endif
	cmd->sg = cmd->se_cmd.t_data_sg = sgtmp;
	cmd->sg_cnt = cmd->se_cmd.t_data_nents = 1;
}

extern struct qs_interface_cbs icbs;
static int
qla_end_notify(struct qsio_immed_notify *notify)
{
	struct fc_cmd *fc_cmd = notify_cmd(notify);
	struct qla_tgt_mgmt_cmd *mcmd = fc_cmd->tgt_cmd;

	mcmd->fc_tm_rsp = notify->notify_status;
	qlt_xmit_tm_rsp(mcmd);
	return 0;
}

void
cmd_sg_free(struct qla_tgt_cmd *cmd)
{
	if (cmd->sg) {
		kfree(cmd->sg);
		cmd->sg = cmd->se_cmd.t_data_sg = NULL;
		cmd->sg_cnt = cmd->se_cmd.t_data_nents = 0;
	}
}

static inline char *
se_data_dir(int dir)
{
	switch(dir) {
	case DMA_NONE:
		return "dma_none";
	case DMA_FROM_DEVICE:
		return "dma_from_device";
	case DMA_TO_DEVICE:
		return "dma_to_device";
	default:
		break;
	}
	return "unknown";
}

void
qla_sc_session_remove_cmd(struct fc_cmd *fc_cmd)
{
	struct qla_tgt_cmd *cmd = fc_cmd->tgt_cmd;
	struct fc_port *sess = cmd->sess;
	struct se_session *se_sess = sess->se_sess;
	struct fc_sess *fc_sess = (struct fc_sess *)se_sess->fabric_sess_ptr;

	fc_sess_remove_cmd(fc_sess, fc_cmd);
}

void
qla_handle_incomp_ctio(struct qsio_scsiio *ctio)
{
	struct fc_cmd *fc_cmd = ctio_cmd(ctio);
	struct qla_tgt_cmd *cmd = fc_cmd->tgt_cmd;

	ctio->ccb_h.flags |= QSIO_DONE_SEND_STATUS;

	DEBUG_INFO_NEW("Aborting incomp ctio cdb %x i_prt %llx t_prt %llx\n", ctio->cdb[0], (unsigned long long)ctio->i_prt[0], (unsigned long long)ctio->t_prt[0]);
	__ctio_free_all(ctio, ctio->local_pool);
	qla_sc_session_remove_cmd(fc_cmd);
	cmd_sg_free(cmd);
	qlt_free_cmd(cmd);
}

static int
qla_end_ctio(struct qsio_scsiio *ctio)
{
	int retval;
	struct ccb_list ctio_list;
	struct fc_cmd *fc_cmd = ctio_cmd(ctio);
	struct qla_tgt_cmd *cmd = fc_cmd->tgt_cmd;
	struct se_cmd *se_cmd = &cmd->se_cmd;

	STAILQ_INIT(&ctio_list);
	ctio->ccb_h.flags |= QSIO_DONE_SEND_STATUS;
	if ((ctio->ccb_h.flags & QSIO_SEND_STATUS) && ctio->ccb_h.target_lun && ctio->ccb_h.tdevice) {
		(*icbs.device_remove_ctio)(ctio, &ctio_list);
	}

	if ((ctio->ccb_h.flags & QSIO_CTIO_ABORTED)) {
		if ((ctio->ccb_h.flags & QSIO_SEND_ABORT_STATUS)) {
			__ctio_free_data(ctio);
			ctio_construct_sense(ctio, SSD_CURRENT_ERROR, SSD_KEY_ABORTED_COMMAND, 0, COMMANDS_CLEARED_BY_ANOTHER_INITIATOR_ASC, COMMANDS_CLEARED_BY_ANOTHER_INITIATOR_ASCQ);
		}
		else {
			DEBUG_INFO_NEW("Aborting incomp ctio cdb %x i_prt %llx t_prt %llx\n", ctio->cdb[0], (unsigned long long)ctio->i_prt[0], (unsigned long long)ctio->t_prt[0]);
			__ctio_free_all(ctio, ctio->local_pool);
			qla_sc_session_remove_cmd(fc_cmd);
			cmd_sg_free(cmd);
			qlt_free_cmd(cmd);
			goto send;
		}
	}

	DEBUG_INFO("cmd %x flags %x\n", ctio->cdb[0], ctio->ccb_h.flags);
	if (ctio->pglist_cnt)
		ctio_sglist_map(ctio, cmd);
	else if (ctio->dxfer_len)
		ctio_buffer_map(ctio, cmd);
	else {
		DEBUG_BUG_ON(cmd->sg);
		cmd_sg_free(cmd);
	}

	cmd->dma_data_direction = ctio_direction(ctio);
	if (fc_cmd->data_dir == DMA_FROM_DEVICE) {
		if (fc_cmd->orig_length > ctio->dxfer_len) {
			se_cmd->residual_count = fc_cmd->orig_length - ctio->dxfer_len;
			se_cmd->se_cmd_flags |= SCF_UNDERFLOW_BIT;
		}
		else if (fc_cmd->orig_length < ctio->dxfer_len) {
			se_cmd->residual_count = ctio->dxfer_len - fc_cmd->orig_length;
			se_cmd->se_cmd_flags |= SCF_OVERFLOW_BIT;
		}
	}

	if (ctio->sense_len)
		memcpy(cmd->sense_buffer, ctio->sense_data, ctio->sense_len);

	DEBUG_INFO("cmd %x target_lun %llu data_dir %s orig_length %u dxfer_len %d pglist_cnt %d sense_len %d residual count %u scsi_status %d\n", ctio->cdb[0], (unsigned long long)ctio->ccb_h.target_lun, se_data_dir(se_cmd->data_dir), se_cmd->orig_length, ctio->dxfer_len, ctio->pglist_cnt, ctio->sense_len, se_cmd->residual_count, ctio->scsi_status);

	cmd->bufflen = se_cmd->data_length = ctio->dxfer_len;
	cmd->se_cmd.scsi_status = ctio->scsi_status;

again:
	if (cmd->sess && cmd->sess->deleted) {
		DEBUG_INFO_NEW("Sess deleted ctio cdb %x i_prt %llx t_prt %llx\n", ctio->cdb[0], (unsigned long long)ctio->i_prt[0], (unsigned long long)ctio->t_prt[0]);
		__ctio_free_all(ctio, ctio->local_pool);
		qla_sc_session_remove_cmd(fc_cmd);
		cmd_sg_free(cmd);
		qlt_free_cmd(cmd);
		goto send;
	}

	fc_cmd->jiffies_at_send = get_jiffies_64();
	if (ctio->dxfer_len && cmd->dma_data_direction == DMA_TO_DEVICE &&
	    cmd->state != QLA_TGT_STATE_DATA_IN) {
		retval = qlt_rdy_to_xfer(cmd);
	}
	else {
		retval = qlt_xmit_response(cmd, QLA_TGT_XMIT_DATA|QLA_TGT_XMIT_STATUS,
					cmd->se_cmd.scsi_status);
	}

	if (unlikely(retval != 0)) {
		if (retval == -EAGAIN) {
			msleep(1);
			goto again;
		}
		DEBUG_INFO_NEW("Aborting incomp ctio cdb %x i_prt %llx t_prt %llx\n", ctio->cdb[0], (unsigned long long)ctio->i_prt[0], (unsigned long long)ctio->t_prt[0]);
		__ctio_free_all(ctio, ctio->local_pool);
		qla_sc_session_remove_cmd(fc_cmd);
		cmd_sg_free(cmd);
		qlt_free_cmd(cmd);;
	}
send:
	if (!STAILQ_EMPTY(&ctio_list))
		(*icbs.device_queue_ctio_list)(&ctio_list);
	return 0;
}

void
qla_end_ccb(void *ccb_void)
{
	struct qsio_hdr *ccb_h = ccb_void;

	if (ccb_h->flags & QSIO_TYPE_CTIO)
	{
		qla_end_ctio((struct qsio_scsiio *)(ccb_h));
	}
	else if (ccb_h->flags & QSIO_TYPE_NOTIFY)
	{
		qla_end_notify((struct qsio_immed_notify *)(ccb_h));
	}
	else
	{
		DEBUG_BUG_ON(1);
	}
}

static void 
qla_sc_handle_data(struct qla_tgt_cmd *cmd)
{
	struct fc_cmd *fc_cmd = (struct fc_cmd *)cmd->se_cmd.priv;
	struct qsio_scsiio *ctio = (struct qsio_scsiio *)fc_cmd->ccb;
 	struct fcq *fcq;

	if (fc_cmd->jiffies_at_send)
		ctio->dxfer_ticks += ticks_delta(fc_cmd->jiffies_at_send, get_jiffies_64());
	cmd_sg_free(cmd);
	ctio->ccb_h.flags &= ~QSIO_DIR_OUT;
 	fcq = ctio_fcq(ctio);
	DEBUG_BUG_ON(!fcq);
	fcq_insert_cmd(fcq, fc_cmd);
}

static void 
qla_sc_free_cmd(struct qla_tgt_cmd *cmd)
{
	struct fc_cmd *fc_cmd = (struct fc_cmd *)cmd->se_cmd.priv;
	struct qsio_scsiio *ctio = (struct qsio_scsiio *)fc_cmd->ccb;
	struct fcq *fcq;
	struct lport *lport;

	if (ctio) {
		if (fc_cmd->jiffies_at_send)
			ctio->dxfer_ticks += ticks_delta(fc_cmd->jiffies_at_send, get_jiffies_64());
		fc_cmd->need_free = 1;
		fcq = ctio_fcq(ctio);
		lport = cmd->sess->tgt->vha->vha_tgt.target_lport_ptr;
		fcq_insert_cmd(fcq ? fcq : lport->fcbridge->fcq, fc_cmd);
		return;
	}

	qla_sc_session_remove_cmd(fc_cmd);
	cmd_sg_free(cmd);
	qlt_free_cmd(cmd);
}

static void
qla_sc_free_mcmd(struct qla_tgt_mgmt_cmd *mcmd)
{
	struct se_cmd *se_cmd;
	struct fc_cmd *fc_cmd;

	if (!mcmd)
		return;

	se_cmd = &mcmd->se_cmd;
	fc_cmd = se_cmd->priv;
	if (fc_cmd)
		mempool_free(fc_cmd, fc_cmd_pool);
	qlt_free_mcmd(mcmd);
}

static int
qla_sc_handle_tmr(struct qla_tgt_mgmt_cmd *mcmd, uint64_t lun,
		uint16_t tmr_func, uint32_t tag)
{
	struct se_cmd *se_cmd = &mcmd->se_cmd;
	struct fc_cmd *fc_cmd;
	struct lport *lport;

	fc_cmd = mempool_alloc(fc_cmd_pool, GFP_ATOMIC);
	if (!fc_cmd) {
		DEBUG_WARN_NEW("Cannot allocate a new fc_cmd\n");
		return -1;
	}
	memset(fc_cmd, 0, sizeof(*fc_cmd));

	se_cmd->priv = fc_cmd;
	fc_cmd->type = QLA_HDR_TYPE_NOTIFY;
	fc_cmd->notify_fn = tmr_func;
	fc_cmd->tgt_cmd = mcmd;
	lport = mcmd->sess->tgt->vha->vha_tgt.target_lport_ptr;
	fcq_insert_cmd(lport->fcbridge->fcq, fc_cmd);
	return 0;
}

static int
qla_sc_handle_cmd(scsi_qla_host_t *vha, struct qla_tgt_cmd *cmd,
		unsigned char *cdb, uint32_t data_length, int fcp_task_attr,
		int data_dir, int bidi)
{
	struct se_cmd *se_cmd = &cmd->se_cmd;
	struct se_session *se_sess = se_cmd->se_sess;
	struct fc_sess *fc_sess;
	struct fc_cmd *fc_cmd;
	struct fcq *fcq;
	struct lport *lport;
	int cdb_size = scsi_command_size(cdb);

	if (!se_sess) {
		DEBUG_WARN_NEW("se sess is NULL\n");
		return -1;
	}

	fc_sess = se_sess->fabric_sess_ptr;
	if (!fc_sess) {
		DEBUG_WARN_NEW("fc sess is NULL\n");
		return -1;
	}

	fc_cmd = mempool_alloc(fc_cmd_pool, GFP_ATOMIC);
	if (!fc_cmd) {
		DEBUG_WARN_NEW("Cannot allocate a new fc_cmd\n");
		return -1;
	}
	memset(fc_cmd, 0, sizeof(*fc_cmd));

	se_cmd->priv = fc_cmd;
	fc_cmd->type = QLA_HDR_TYPE_CTIO;
	fc_cmd->tgt_cmd = cmd;
	fc_cmd->orig_length = data_length;
	fc_cmd->task_attr = fcp_task_attr;
	fc_cmd->data_dir = data_dir;
	memcpy(fc_cmd->cdb, cdb, 16);
	if (cdb_size > 16) {
		DEBUG_WARN_NEW("Received > 16byte command %x\n", cdb[0]);
		cdb_size = 16;
	}
	fc_sess_add_cmd(fc_sess, fc_cmd);
	lport = cmd->sess->tgt->vha->vha_tgt.target_lport_ptr;
	fcq = fcq_get(lport->fcbridge, cmd->unpacked_lun); 
	fcq_insert_cmd(fcq, fc_cmd);
	return 0;
}

static void
qla_sc_free_session(struct fc_port *sess)
{
	struct se_session *se_sess;
	struct fc_sess *fc_sess;
	uint64_t i_prt[2], t_prt[2];
	struct lport *lport;

	se_sess = sess->se_sess;
	if (!se_sess)
		return;

	se_sess = sess->se_sess;
	fc_sess = se_sess->fabric_sess_ptr;
	lport = sess->tgt->vha->vha_tgt.target_lport_ptr;
	DEBUG_INFO_NEW("free session for lport %d %8phC pending cmds %d loop_id %d key %x\n", lport->id, fc_sess->wwpn, atomic_read(&fc_sess->pending_cmds), fc_sess->loop_id, fc_sess->key);
	while (atomic_read(&fc_sess->pending_cmds) > 0)
		msleep(100);

	i_prt[0] = wwn_to_u64(sess->port_name);
	i_prt[1] = 0;
	t_prt[0] = wwn_to_u64(sess->tgt->vha->port_name);
	t_prt[1] = 0;
	fcbridge_free_initiator(i_prt, t_prt);
	fc_sess_remove(lport, fc_sess);
	kmem_cache_free(se_session_cache, se_sess);
}

static int
qla_sc_check_initiator_node_acl( scsi_qla_host_t *vha, unsigned char *fc_wwpn,
	struct fc_port *sess)
{
	struct lport *lport;
	struct fc_sess *fc_sess;
	struct se_session *se_sess;
	uint16_t loop_id;
	be_id_t s_id;
	uint32_t key;
	unsigned long flags;

	lport = vha->vha_tgt.target_lport_ptr;
	if (!lport) {
		DEBUG_WARN_NEW("lport is NULL\n");
		return -EINVAL;
	}

	loop_id = sess->loop_id;
	if (lport->loop_sess_list[loop_id]) {
		DEBUG_WARN_NEW("lport %d already has a fc_sess at loop id %d\n", lport->id, loop_id);
		return -EINVAL;
	}

	fc_sess = fc_sess_new(lport);
	if (!fc_sess)
		return -ENOMEM;

        se_sess = kmem_cache_zalloc(se_session_cache, GFP_ATOMIC);
        if (!se_sess) {
		kmem_cache_free(fc_sess_cache, fc_sess);
		return -ENOMEM;
	}

	s_id = port_id_to_be_id(sess->d_id);
	key = sid_to_key(s_id);
	fc_sess->sess = sess;
	fc_sess->key = key;
	fc_sess->loop_id = loop_id;
	fc_sess->active = 1;
	memcpy(fc_sess->wwpn, fc_wwpn, WWN_SIZE);

	spin_lock_irqsave(&lport->lport_lock, flags);
	TAILQ_INSERT_HEAD(&lport->active_sess_list, fc_sess, a_list);
	TAILQ_INSERT_HEAD(&lport->sess_list, fc_sess, s_list);

	se_sess->fabric_sess_ptr = fc_sess;
	((struct fc_port *)sess)->se_sess = se_sess;
	lport->loop_sess_list[loop_id] = fc_sess;
	spin_unlock_irqrestore(&lport->lport_lock, flags);

	DEBUG_INFO_NEW("new session for lport %d %8phC loop_id %d key %x\n", lport->id, fc_sess->wwpn, loop_id, key);
	return 0;
}

static struct fc_port *
qla_sc_find_sess_by_s_id(scsi_qla_host_t *vha, const be_id_t s_id)
{
	struct lport *lport;
	struct fc_sess *iter;
	struct fc_port *sess = NULL;
	unsigned long flags;
	uint32_t key = sid_to_key(s_id);

	lport = vha->vha_tgt.target_lport_ptr;
	DEBUG_BUG_ON(!lport);
	if (!lport) {
		DEBUG_WARN_NEW("Cannot find lport ptr\n");
		return NULL;
	}
	spin_lock_irqsave(&lport->lport_lock, flags);
	TAILQ_FOREACH(iter, &lport->active_sess_list, a_list) {
		if (iter->key == key) {
			sess = iter->sess;
			if (iter != TAILQ_FIRST(&lport->active_sess_list)) {
				TAILQ_REMOVE(&lport->active_sess_list, iter, a_list);
				TAILQ_INSERT_HEAD(&lport->active_sess_list, iter, a_list);
			}
			break;
		}
	}
	spin_unlock_irqrestore(&lport->lport_lock, flags);
	if (!sess)
		DEBUG_INFO_NEW("cannot find sess for lport %d key %x\n", lport->id, key);
	return sess;
}

static struct fc_port *
qla_sc_find_sess_by_loop_id(scsi_qla_host_t *vha, const uint16_t loop_id)
{
	struct lport *lport;
	struct fc_sess *fc_sess;
	unsigned long flags;

	lport = vha->vha_tgt.target_lport_ptr;
	DEBUG_BUG_ON(!lport);
	if (!lport) {
		DEBUG_WARN_NEW("Cannot find lport ptr\n");
		return NULL;
	}
	spin_lock_irqsave(&lport->lport_lock, flags);
	fc_sess = lport->loop_sess_list[loop_id];
	spin_unlock_irqrestore(&lport->lport_lock, flags);
	if (!fc_sess) {
		DEBUG_INFO_NEW("cannot find sess for lport %d loop_id %d\n", lport->id, loop_id);
		return NULL;
	}
	return fc_sess->sess;
}

static void
qla_sc_clear_nacl_from_fcport_map(struct fc_port *sess)
{
	struct se_session *se_sess = sess->se_sess;
	struct fc_sess *fc_sess = se_sess->fabric_sess_ptr;
	struct lport *lport;
	unsigned long flags;

	lport = sess->tgt->vha->vha_tgt.target_lport_ptr;
	DEBUG_INFO_NEW("clear for lport %d %8phC pending cmds %d loop_id %d key %x\n", lport->id, fc_sess->wwpn, atomic_read(&fc_sess->pending_cmds), fc_sess->loop_id, fc_sess->key);
	spin_lock_irqsave(&lport->lport_lock, flags);
	if (fc_sess->active) {
		lport->loop_sess_list[fc_sess->loop_id] = NULL;
		TAILQ_REMOVE(&lport->active_sess_list, fc_sess, a_list);
		fc_sess->active = 0;
	}
	spin_unlock_irqrestore(&lport->lport_lock, flags);
	return;
}

static void
qla_sc_release_session(struct kref *kref)
{
	struct fc_port *sess = container_of(kref, struct fc_port, sess_kref);

	sess->explicit_logout = 1;
	qlt_unreg_sess(sess);
}

static void
qla_sc_put_sess(struct fc_port *sess)
{
	if (!sess)
		return;

	kref_put(&sess->sess_kref, qla_sc_release_session);
}

static void
qla_sc_shutdown_sess(struct fc_port *sess)
{
}

static void qla_sc_update_sess(struct fc_port *sess, port_id_t s_id,
				    uint16_t loop_id, bool conf_compl_supported)
{
	struct se_session *se_sess = sess->se_sess;
	struct fc_sess *fc_sess = se_sess->fabric_sess_ptr;
	struct lport *lport;
	unsigned long flags;
	uint16_t old_loop_id;
	u32 key, old_key;

	DEBUG_INFO_NEW("update sess d_id for loop id %u\n", loop_id);
	key = (((u32) s_id.b.domain << 16) |
	       ((u32) s_id.b.area   <<  8) |
	       ((u32) s_id.b.al_pa));

	lport = sess->tgt->vha->vha_tgt.target_lport_ptr;
	old_key = fc_sess->key;
	old_loop_id = fc_sess->loop_id;
	spin_lock_irqsave(&lport->lport_lock, flags);
	if (lport->loop_sess_list[loop_id]) {
		DEBUG_WARN_NEW("lport %d already has a fc_sess at loop id %d\n", lport->id, loop_id);
		spin_unlock_irqrestore(&lport->lport_lock, flags);
		return;
	}

	lport->loop_sess_list[old_loop_id] = NULL;
	lport->loop_sess_list[loop_id] = fc_sess;
	fc_sess->loop_id = loop_id;
	fc_sess->key = key;
	sess->d_id = s_id;
	sess->loop_id = loop_id;
	sess->conf_compl_supported = conf_compl_supported;
	spin_unlock_irqrestore(&lport->lport_lock, flags);
}

static struct qla_tgt_cmd *
qla_sc_find_cmd_by_tag(struct fc_port *sess, uint64_t tag)
{
	struct se_session *se_sess;
	struct fc_sess *fc_sess;
	unsigned long flags;
	struct qla_tgt_cmd *cmd, *ret = NULL;
	struct fc_cmd *iter;

	if (!sess->se_sess)
		return NULL;

	se_sess = sess->se_sess;
	fc_sess = (struct fc_sess *)se_sess->fabric_sess_ptr;
	spin_lock_irqsave(&fc_sess->sess_lock, flags);
	TAILQ_FOREACH(iter, &fc_sess->cmd_list, f_list) {
		if (iter->type != QLA_HDR_TYPE_CTIO)
			continue;
		cmd = iter->tgt_cmd; 
		if (cmd->se_cmd.tag == tag) {
			ret = cmd;
			break;
		}
	}
	spin_unlock_irqrestore(&fc_sess->sess_lock, flags); 
	return ret;
}

static int
qla_sc_get_dif_tags(struct qla_tgt_cmd *cmd, uint16_t *pfw_prot_opts)
{
	return 0;
}

static int
qla_sc_chk_dif_tags(uint32_t tag)
{
	return 0;
}

static struct qla_tgt_cmd *
qla_sc_get_cmd(struct fc_port *sess)
{
	struct qla_tgt_cmd *cmd;

	cmd = kmem_cache_zalloc(tgt_cmd_cache, GFP_ATOMIC);
	cmd->se_cmd.se_sess = sess->se_sess;
	return cmd;
}

static void
qla_sc_rel_cmd(struct qla_tgt_cmd *cmd)
{
	struct fc_cmd *fc_cmd = (struct fc_cmd *)cmd->se_cmd.priv;

	if (fc_cmd)
		mempool_free(fc_cmd, fc_cmd_pool);
	kmem_cache_free(tgt_cmd_cache, cmd);
}

int
fcbridge_i_prt_valid(struct fcbridge *fcbridge, uint64_t i_prt[])
{
	scsi_qla_host_t *ha = fcbridge->ha;

	if (i_prt[1] || i_prt[0] != wwn_to_u64(ha->port_name))
		return 1;
	else
		return 0;
}

void
fcbridge_get_tport(struct fcbridge *fcbridge, uint64_t wwpn[])
{
	wwpn[0] = wwn_to_u64(fcbridge->ha->port_name);
	wwpn[1] = 0;
}

static int
lport_register_cb(struct scsi_qla_host *vha, void *lportptr, uint64_t npiv_wwpn, uint64_t npiv_wwnn)
{
	struct qla_hw_data *ha = vha->hw;
	struct lport *lport = (struct lport *)lportptr;
	struct fcbridge *fcbridge;

	fcbridge = fcbridge_new(vha, vha->host_no);
	if (!fcbridge) {
		DEBUG_WARN_NEW("cannot create fcbridge for host no %lu\n", vha->host_no);
		return -1;
	}

	lport->vha = vha;
	lport->fcbridge = fcbridge;
	ha->tgt.tgt_ops = &qla_sc_template;
	vha->vha_tgt.target_lport_ptr = lportptr;
	return 0;
}

static struct lport *
lport_alloc(void)
{
	struct lport *lport;

	lport = kmem_cache_zalloc(lport_cache, GFP_KERNEL);
	if (!lport)
		return NULL;
	spin_lock_init(&lport->lport_lock);
	TAILQ_INIT(&lport->sess_list);
	TAILQ_INIT(&lport->active_sess_list);
	return lport;
}

uint32_t lport_id;
static int
__qla_sc_hba_enable(struct hba_info *hba_info)
{
	struct lport *iter, *lport = NULL;
	int retval, mode = hba_info->enable;
	scsi_qla_host_t *vha;
	struct qla_tgt *tgt;
	struct fcbridge *fcbridge;
	unsigned long flags;

	spin_lock_irqsave(&lport_list_lock, flags);
	TAILQ_FOREACH(iter, &lport_list, l_list) {
		if (iter->wwpn == hba_info->wwpn) {
			lport = iter;
			break;
		}
	}
	spin_unlock_irqrestore(&lport_list_lock, flags);

	if (!lport) {
		lport = lport_alloc();
		if (!lport) {
			DEBUG_WARN_NEW("lport allocation failure\n");
			return -1;
		}
		lport->wwpn = hba_info->wwpn;
		retval = qlt_lport_register(lport, lport->wwpn, 0, 0, lport_register_cb);
		if (retval != 0) { 
			DEBUG_WARN_NEW("Cannot register lbport for wwpn %llx\n", (unsigned long long)lport->wwpn);
			kmem_cache_free(lport_cache, lport);
			return -1;
		}

		spin_lock_irqsave(&lport_list_lock, flags);
		lport->id = lport_id++;
		TAILQ_INSERT_HEAD(&lport_list, lport, l_list);
		spin_unlock_irqrestore(&lport_list_lock, flags);
	}

       	vha = lport->vha;
       	tgt = vha->vha_tgt.qla_tgt;
	fcbridge = lport->fcbridge;

	DEBUG_INFO_NEW("wwpn %llx mode %d fcbridge target mode %d initiator mode %d\n", (unsigned long long)lport->wwpn, mode, fcbridge->target_mode, fcbridge->initiator_mode);
	switch (mode) {
	case FC_INITIATOR_MODE:
		if (!fcbridge->target_mode) {
			DEBUG_INFO_NEW("initiator mode already enabled\n");
			return 0;
		}
		fcbridge->target_mode = 0;
		fcbridge->initiator_mode = 1;
		if (tgt->tgt_stopped)
			return 0;
		qlt_stop_phase1(tgt);
		qlt_stop_phase2(tgt);
		return 0;
	case FC_TARGET_MODE:
		if (fcbridge->target_mode && !fcbridge->initiator_mode) {
			DEBUG_INFO_NEW("target mode already enabled and initiator mode disabled\n");
			return 0;
		}
		fcbridge->target_mode = 1;
		fcbridge->initiator_mode = 0;
		qlt_enable_vha(vha);
		return 0;
	case FC_DUAL_MODE:
		if (fcbridge->target_mode && fcbridge->initiator_mode) {
			DEBUG_INFO_NEW("dual target and initator mode already enabled\n");
			return 0;
		}

		fcbridge->target_mode = 1;
		qlt_enable_vha(vha);
		fcbridge->initiator_mode = 1;
		return 0;
	default:
		return -1;
	}
	return 0;
}

static struct qla_tgt_func_tmpl qla_sc_template = {
	.find_cmd_by_tag	= qla_sc_find_cmd_by_tag,
	.handle_cmd		= qla_sc_handle_cmd,
	.handle_data		= qla_sc_handle_data,
	.handle_tmr		= qla_sc_handle_tmr,
	.get_cmd		= qla_sc_get_cmd,
	.rel_cmd		= qla_sc_rel_cmd,
	.free_cmd		= qla_sc_free_cmd,
	.free_mcmd		= qla_sc_free_mcmd,
	.free_session		= qla_sc_free_session,
	.update_sess		= qla_sc_update_sess,
	.check_initiator_node_acl	= qla_sc_check_initiator_node_acl,
	.find_sess_by_s_id	= qla_sc_find_sess_by_s_id,
	.find_sess_by_loop_id	= qla_sc_find_sess_by_loop_id,
	.clear_nacl_from_fcport_map	= qla_sc_clear_nacl_from_fcport_map,
	.put_sess		= qla_sc_put_sess,
	.shutdown_sess		= qla_sc_shutdown_sess,
	.get_dif_tags		= qla_sc_get_dif_tags,
	.chk_dif_tags		= qla_sc_chk_dif_tags,
};

extern void fcbridge_hba_debug(struct hba_info *);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0))
long qla_sc_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#else
int qla_sc_ioctl(struct inode *i, struct file *f, uint32_t cmd, unsigned long arg)
#endif
{
	struct hba_info hba_info;
	void __user *userp = (void __user *)arg;
	int err = 0;
	int retval = 0;

	if (!capable(CAP_SYS_ADMIN)) {
		DEBUG_WARN_NEW("not sysadmin\n");
		return (-EPERM);
	}

	if (_IOC_TYPE (cmd) != FC_MAGIC)
		return -ENOTTY;

	if (_IOC_DIR (cmd) & _IOC_READ) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 1)
		err = !access_ok (userp, _IOC_SIZE (cmd));
#else
		err = !access_ok (VERIFY_READ, userp, _IOC_SIZE (cmd));
#endif
	}

	if (_IOC_DIR (cmd) & _IOC_WRITE) {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5,0,0)) || (defined(RHEL_MAJOR) && RHEL_MAJOR >= 8 && RHEL_MINOR >= 1)
		err = !access_ok (userp, _IOC_SIZE (cmd));
#else
		err = !access_ok (VERIFY_WRITE, userp, _IOC_SIZE (cmd));
#endif
	}

	if (err)
		return -EPERM;

	switch (cmd) {
	case FCTARGIOCENABLEHBA:
		if ((retval = copyin(userp, &hba_info, sizeof(hba_info))) != 0)
			break;
		retval = __qla_sc_hba_enable(&hba_info);
		break;
	case FCTARGIOCHBADEBUG:
		if ((retval = copyin(userp, &hba_info, sizeof(hba_info))) != 0)
			break;
		fcbridge_hba_debug(&hba_info);
		break;
	}
	return retval;
}

static struct file_operations qla_sc_fops = {
	.owner = THIS_MODULE,
	.llseek = noop_llseek,
	.compat_ioctl = qla_sc_ioctl,
	.unlocked_ioctl = qla_sc_ioctl,
};

static int dev_major;

static void
exit_caches(void)
{
	if (fc_sess_cache) {
		kmem_cache_destroy(fc_sess_cache);
		fc_sess_cache = NULL;
	}

	if (fc_cmd_pool) {
		mempool_destroy(fc_cmd_pool);
		fc_cmd_pool = NULL;
	}

	if (fc_cmd_cache) {
		kmem_cache_destroy(fc_cmd_cache);
		fc_cmd_cache = NULL;
	}

	if (tgt_cmd_cache) {
		kmem_cache_destroy(tgt_cmd_cache);
		tgt_cmd_cache = NULL;
	}

	if (lport_cache) {
		kmem_cache_destroy(lport_cache);
		lport_cache = NULL;
	}

	if (se_session_cache) {
		kmem_cache_destroy(se_session_cache);
		se_session_cache = NULL;
	}
}

static void *
fc_cmd_slab_alloc(gfp_t gfp_mask, void *cache)
{
	return kmem_cache_alloc(fc_cmd_cache, gfp_mask);
}

static void
fc_cmd_slab_free(void *element, void *cache)
{
	return kmem_cache_free(fc_cmd_cache, element);
}

static int __init
qla_sc_init(void)
{

	fc_sess_cache = kmem_cache_create("fc_sess_cache", sizeof(struct fc_sess), 0, 0, NULL);
	if (!fc_sess_cache)
		return -ENOMEM;

	fc_cmd_cache = kmem_cache_create("fc_cmd_cache", sizeof(struct fc_cmd), 0, 0, NULL);
	if (!fc_cmd_cache) {
		exit_caches();
		return -ENOMEM;
	}

	fc_cmd_pool = mempool_create(1024, fc_cmd_slab_alloc, fc_cmd_slab_free, fc_cmd_cache);
	if (!fc_cmd_pool) {
		exit_caches();
		return -ENOMEM;
	}

	tgt_cmd_cache = kmem_cache_create("tc_cmd_cache", sizeof(struct qla_tgt_cmd), 0, 0, NULL);
	if (!tgt_cmd_cache) {
		exit_caches();
		return -ENOMEM;
	}

	lport_cache = kmem_cache_create("lport_cache", sizeof(struct lport), 0, 0, NULL);
	if (!tgt_cmd_cache) {
		exit_caches();
		return -ENOMEM;
	}

	se_session_cache = kmem_cache_create("se_session_cache", sizeof(struct se_session), 0, 0, NULL);
	if (!tgt_cmd_cache) {
		exit_caches();
		return -ENOMEM;
	}

	dev_major = register_chrdev(0, FC_DEV_NAME, &qla_sc_fops);
	if (unlikely(dev_major < 0)) {
		exit_caches();
		return -(EIO);
	}

	return 0;
}

static void __exit
qla_sc_exit(void)
{
	struct lport *lport;
	struct fcbridge *fcbridge;
	struct qla_tgt *tgt;

	while ((lport = TAILQ_FIRST(&lport_list)) != NULL) {
		TAILQ_REMOVE(&lport_list, lport, l_list);
       		tgt = lport->vha->vha_tgt.qla_tgt;
		fcbridge = lport->fcbridge;
		if (!tgt->tgt_stopped) {
			qlt_stop_phase1(lport->vha->vha_tgt.qla_tgt);
			qlt_stop_phase2(lport->vha->vha_tgt.qla_tgt);
		}
		qlt_lport_deregister(lport->vha);
		fcbridge_exit(fcbridge);
		kmem_cache_free(lport_cache, lport);
	}

	unregister_chrdev(dev_major, FC_DEV_NAME);
	exit_caches();
}

MODULE_VERSION("1.0");
MODULE_DESCRIPTION("FC interface driver");
MODULE_LICENSE("GPL");
module_init(qla_sc_init);
module_exit(qla_sc_exit);

