/*
 * Copyright (C) 2002-2003 Ardis Technolgies <roman@ardistech.com>
 *
 * Released under the terms of the GNU GPL v2.0.
 */

#include <ctype.h>
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <sys/stat.h>
#include <errno.h>

#include "iscsid.h"

static struct session *session_alloc(u32 tid)
{
	struct session *session;
	struct target *target = target_find_by_id(tid);

	if (!target)
		return NULL;
	if (!(session = malloc(sizeof(*session))))
		return NULL;
	memset(session, 0, sizeof(*session));

	session->target = target;
	INIT_LIST_HEAD(&session->slist);
	session->timestamp = (unsigned long)time(NULL);
	insque(&session->slist, &target->sessions_list);

	return session;
}

struct session *session_find_name(u32 tid, const char *iname, union iscsi_sid sid)
{
	struct session *session;
	struct target *target;

	if (!(target = target_find_by_id(tid)))
		return NULL;

	log_debug(1, "session_find_name: %s %#" PRIx64, iname, sid.id64);
	list_for_each_entry(session, &target->sessions_list, slist) {
		if (!memcmp(sid.id.isid, session->sid.id.isid, 6) &&
		    !strcmp(iname, session->initiator))
			return session;
	}

	return NULL;
}

struct session *session_find_id(u32 tid, u64 sid)
{
	struct session *session;
	struct target *target;

	if (!(target = target_find_by_id(tid)))
		return NULL;

	log_debug(1, "session_find_id: %#" PRIx64, sid);
	list_for_each_entry(session, &target->sessions_list, slist) {
		if (session->sid.id64 == sid)
			return session;
	}

	return NULL;
}

int session_exist(u32 t_tid, u64 t_sid)
{
	struct session_info info;

	memset(&info, 0x0, sizeof(info));

	info.tid = t_tid;
	info.sid = t_sid;

	return !ki->session_info(&info);
}

int
target_get_initiator_id(struct target *target, char *initiator)
{
	char *new_initiator;
	int i, start, max_id;

	for (i = 0; i < MAX_INITIATORS_PER_TARGET; i++) {
		if (!target->initiators[i])
			continue;
		if (strcmp(target->initiators[i], initiator) == 0)
			return i;
	}

	new_initiator = strdup(initiator);
	if (!new_initiator)
		return -1;

	start = target->last_init_id;
	max_id = MAX_INITIATORS_PER_TARGET;
retry:
	for (i = start; i < max_id; i++) {
		if (target->initiators[i])
			continue;
		target->initiators[i] = new_initiator;
		target->last_init_id = i+1;
		if (target->last_init_id == MAX_INITIATORS_PER_TARGET)
			target->last_init_id = 0;
		return i;
	}
	if (!start) {
		free(new_initiator);
		return -1;
	}
	start = 0;
	max_id = target->last_init_id;
	goto retry;
}

int session_create(struct connection *conn)
{
	struct target *target;
	struct session *session;
	char *initiator;
	int err, init_id, i;

	if (!conn->session) {
		initiator = strdup(conn->initiator);
		if (!initiator)
			return -ENOMEM;

		session = session_alloc(conn->tid);
		if (!session) {
			free(initiator);
			return -ENOMEM;
		}

		target = session->target;
		init_id = target_get_initiator_id(target, initiator);
		if (init_id < 0) {
			log_warning("initiator %s target %u initiator limit reached\n", conn->initiator, conn->tid);
			free(session);
			free(initiator);
			return -ENOMEM;
		}

		session->sid = conn->sid;
		for (i = 1; i <= MAX_CONNS_PER_INITIATOR; i++) {
			session->sid.id.tsih = (init_id << 5 | i);
			if (!session_exist(conn->tid, session->sid.id64))
				break;

		}

		if (i > MAX_CONNS_PER_INITIATOR) {
			log_warning("initiator %s target %u conns %d greater than max %d\n", conn->initiator, conn->tid, i, MAX_CONNS_PER_INITIATOR);
			free(session);
			free(initiator);
			return -ENOMEM;
		}
		log_info("initiator %s target %u tsih %d init_id %d sid %llx\n", conn->initiator, conn->tid, session->sid.id.tsih, init_id, (unsigned long long)session->sid.id64);

		session->initiator = initiator;
		conn->session = session;
		session->conn_cnt = 1;
		conn->sid = session->sid;
	} else {
		session = conn->session;
		conn->sid = session->sid;

		if (session_exist(conn->tid, session->sid.id64))
			return 0;
	}

	log_info("initiator %s session %llx\n", conn->initiator, (unsigned long long)session->sid.id64);

	err = ki->session_create(conn->tid, session->sid.id64, conn->exp_cmd_sn,
			   conn->max_cmd_sn, session->initiator);
	if (err < 0 && err != -EEXIST) {
		log_error("unable to create session %#" PRIx64 " in target %u: %d",
			session->sid.id64, conn->tid, errno);
		goto out;
	}

	err = ki->param_set(conn->tid, session->sid.id64, key_session, 0,
							conn->session_param);
	if (err < 0)
		log_warning("unable to set parameters for session %#" PRIx64 " in target %u: %d",
			session->sid.id64, conn->tid, errno);

	return 0;
out:
	conn->session = NULL;

	if (session->target) {
		remque(&session->slist);
		--session->target->nr_sessions;
	}

	free(session->initiator);
	free(session);

	return err;
}

int session_remove(struct session *session)
{
	int err;

	log_debug(1, "session_remove: %#"  PRIx64, session->sid.id64);

	err = ki->session_destroy(session->target->tid, session->sid.id64);
	if (err < 0 && err != -ENOENT) {
		log_error("unable to delete session %#" PRIx64 " in target %u: %d",
			session->sid.id64, session->target->tid, errno);
		return err;
	}

	if (session->target) {
		remque(&session->slist);
		--session->target->nr_sessions;
	}

	free(session->initiator);
	free(session);

	return 0;
}
